import string

GL_INT = "GL_INT"
GL_FLOAT = "GL_FLOAT"
GL_STRING = "GL_STRING"
GL_IDENTIFIER = "GL_IDENTIFIER"
GL_KEYWORD = "GL_KEYWORD"
GL_PLUS = "GL_PLUS"
GL_MINUS = "GL_MINUS"
GL_MUL = "GL_MUL"
GL_DIV = "GL_DIV"
GL_MOD = "GL_MOD"
GL_FLOORDIV = "GL_FLOORDIV"
GL_POW = "GL_POW"
GL_EQ = "GL_EQ"
GL_LPAREN = "GL_LPAREN"
GL_RPAREN = "GL_RPAREN"
GL_LBRACE = "GL_LBRACE"
GL_RBRACE = "GL_RBRACE"
GL_COMMA = "GL_COMMA"
GL_DOT = "GL_DOT"
GL_EE = "GL_EE"
GL_NE = "GL_NE"
GL_LT = "GL_LT"
GL_GT = "GL_GT"
GL_LTE = "GL_LTE"
GL_GTE = "GL_GTE"
GL_EOF = "GL_EOF"
GL_LSQUARE = "GL_LSQUARE"
GL_RSQUARE = "GL_RSQUARE"
GL_PLUSPLUS = "GL_PLUSPLUS"
GL_MINUSMINUS = "GL_MINUSMINUS"
GL_COLON = "GL_COLON"
GL_QMARK = "GL_QMARK"
GL_PLUSEQ = "PLUSEQ"
GL_MINUSEQ = "MINUSEQ"
GL_MULEQ = "MULEQ"
GL_DIVEQ = "DIVEQ"
GL_POWEQ = "POWEQ"
GL_MODEQ = "MODEQ"
GL_FLOORDIVEQ = "FLOORDIVEQ"
GL_BIT_AND = "GL_BIT_AND"
GL_BIT_OR = "GL_BIT_OR"
GL_BIT_XOR = "GL_BIT_XOR"
GL_BIT_NOT = "GL_BIT_NOT"
GL_LSHIFT = "GL_LSHIFT"
GL_RSHIFT = "GL_RSHIFT"
GL_BIT_ANDEQ = "GL_BIT_ANDEQ"
GL_BIT_OREQ = "GL_BIT_OREQ"
GL_BIT_XOREQ = "GL_BIT_XOREQ"
GL_LSHIFTEQ = "GL_LSHIFTEQ"
GL_RSHIFTEQ = "GL_RSHIFTEQ"

KEYWORDS = [
    "LET",
    "FINAL",
    "PRINT",
    "PRINTLN",
    "IF",
    "THEN",
    "ELSE",
    "ENDIF",
    "AND",
    "OR",
    "NOT",
    "IS",
    "DEF",
    "ENDDEF",
    "RETURN",
    "CLASS",
    "ENDCLASS",
    "INHERITS",
    "NEW",
    "THIS",
    "STATIC",
    "WHILE",
    "ENDWHILE",
    "BREAK",
    "CONTINUE",
    "FOR",
    "IN",
    "ENDFOR",
    "TRY",
    "CATCH",
    "FINALLY",
    "ENDTRY",
    "THROW",
    "SWITCH",
    "CASE",
    "DEFAULT",
    "ENDSWITCH",
    "PUBLIC",
    "PRIVATE",
    "PROTECTED",
    "INSTANCEOF",
    "SUPER",
    "ENUM",
    "ENDENUM",
]

DIGITS = "0123456789"
LETTERS = string.ascii_letters
LETTERS_DIGITS = LETTERS + DIGITS



from .constants import *
from .errors import InvalidSyntaxError
from .nodes import *
from .lexer import Token


class ParseResult:
    def __init__(self):
        self.error = None
        self.node = None
        self.advance_count = 0

    def register_advancement(self):
        self.advance_count += 1

    def register(self, res):
        self.advance_count += res.advance_count
        if res.error:
            self.error = res.error
        return res.node

    def success(self, node):
        self.node = node
        return self

    def failure(self, error):
        if not self.error or self.advance_count == 0:
            self.error = error
        return self


class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.tok_idx = -1
        self.loop_count = 0
        self.advance()

    def advance(self):
        self.tok_idx += 1
        if self.tok_idx < len(self.tokens):
            self.current_tok = self.tokens[self.tok_idx]
        return self.current_tok

    def parse(self):
        res = ParseResult()
        statements = []
        pos_start = self.current_tok.pos_start.copy()

        while self.current_tok.type != GL_EOF:
            if self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
                "ENDDEF",
                "ENDIF",
                "ENDCLASS",
                "ENDWHILE",
                "ENDFOR",
                "ENDENUM",
            ):
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        f"Unexpected '{self.current_tok.value}'",
                    )
                )

            statement = res.register(self.statement())
            if res.error:
                return res
            statements.append(statement)

        return res.success(
            StatementListNode(statements, pos_start, self.current_tok.pos_start.copy())
        )

    def statement_list(self, end_keywords):
        res = ParseResult()
        statements = []
        pos_start = self.current_tok.pos_start.copy()

        while self.current_tok.type != GL_EOF and not (
            self.current_tok.type == GL_KEYWORD
            and self.current_tok.value in end_keywords
        ):
            statements.append(res.register(self.statement()))
            if res.error:
                return res

        return res.success(
            StatementListNode(statements, pos_start, self.current_tok.pos_start.copy())
        )

    def statement(self):
        res = ParseResult()

        if self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
            "PUBLIC",
            "PRIVATE",
            "PROTECTED",
            "FINAL",
        ):
            visibility = "PUBLIC"
            is_final = False

            while self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
                "PUBLIC",
                "PRIVATE",
                "PROTECTED",
                "FINAL",
            ):
                if self.current_tok.value == "FINAL":
                    is_final = True
                else:
                    visibility = self.current_tok.value

                res.register_advancement()
                self.advance()

            if self.current_tok.matches(GL_KEYWORD, "ENUM"):
                enum_node = res.register(self.enum_def())
                if res.error:
                    return res
                enum_node.visibility = visibility
                return res.success(enum_node)

            expr = res.register(self.expr())
            if res.error:
                return res

            if isinstance(expr, (SetAttrNode, VarAssignNode)):
                return res.success(
                    VisibilityStmtNode(visibility, expr, is_final=is_final)
                )

            return res.failure(
                InvalidSyntaxError(
                    expr.pos_start,
                    expr.pos_end,
                    "Visibility modifiers can only be used with variable or attribute assignments",
                )
            )

        if self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
            "PRINT",
            "PRINTLN",
        ):
            is_println = self.current_tok.value == "PRINTLN"
            res.register_advancement()
            self.advance()

            exprs = []
            first_expr = res.register(self.expr())
            if res.error:
                return res
            exprs.append(first_expr)

            while True:
                if self.current_tok.type == GL_COMMA:
                    res.register_advancement()
                    self.advance()
                    exprs.append(res.register(self.expr()))
                    if res.error:
                        return res
                elif self.current_tok.pos_start.ln == exprs[
                    -1
                ].pos_end.ln and self.current_tok.type not in (GL_EOF, GL_KEYWORD):

                    next_expr = res.register(self.expr())
                    if res.error:
                        return res
                    exprs.append(next_expr)
                else:
                    break

            return res.success(PrintNode(exprs, should_newline=is_println))

        if self.current_tok.matches(GL_KEYWORD, "IF"):
            res.register_advancement()
            self.advance()

            cases = []
            else_case = None

            condition = res.register(self.expr())
            if res.error:
                return res

            if not self.current_tok.matches(GL_KEYWORD, "THEN"):
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected 'THEN'",
                    )
                )
            res.register_advancement()
            self.advance()

            body = res.register(self.statement_list(("ELSE", "ENDIF")))
            if res.error:
                return res
            cases.append((condition, body))

            while self.current_tok.matches(GL_KEYWORD, "ELSE"):

                is_else_if = False
                if self.tok_idx + 1 < len(self.tokens):
                    if self.tokens[self.tok_idx + 1].matches(GL_KEYWORD, "IF"):
                        is_else_if = True

                if is_else_if:
                    res.register_advancement()
                    self.advance()
                    res.register_advancement()
                    self.advance()

                    condition = res.register(self.expr())
                    if res.error:
                        return res

                    if not self.current_tok.matches(GL_KEYWORD, "THEN"):
                        return res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected 'THEN'",
                            )
                        )
                    res.register_advancement()
                    self.advance()

                    body = res.register(self.statement_list(("ELSE", "ENDIF")))
                    if res.error:
                        return res
                    cases.append((condition, body))

                else:
                    res.register_advancement()
                    self.advance()

                    else_case = res.register(self.statement_list(("ENDIF",)))
                    if res.error:
                        return res

                    break

            if not self.current_tok.matches(GL_KEYWORD, "ENDIF"):
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected 'ENDIF'",
                    )
                )

            res.register_advancement()
            self.advance()

            return res.success(IfNode(cases, else_case))

        if self.current_tok.matches(GL_KEYWORD, "WHILE"):
            return self.while_expr()

        if self.current_tok.matches(GL_KEYWORD, "FOR"):
            return self.for_expr()

        if self.current_tok.matches(GL_KEYWORD, "BREAK"):
            if self.loop_count == 0:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "'BREAK' outside of loop",
                    )
                )
            pos_start = self.current_tok.pos_start.copy()
            res.register_advancement()
            self.advance()
            return res.success(BreakNode(pos_start, self.current_tok.pos_start.copy()))

        if self.current_tok.matches(GL_KEYWORD, "CONTINUE"):
            if self.loop_count == 0:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "'CONTINUE' outside of loop",
                    )
                )

            pos_start = self.current_tok.pos_start.copy()
            res.register_advancement()
            self.advance()
            return res.success(
                ContinueNode(pos_start, self.current_tok.pos_start.copy())
            )

        if self.current_tok.matches(GL_KEYWORD, "LET"):
            res.register_advancement()
            self.advance()

            if self.current_tok.type == GL_LSQUARE:
                res.register_advancement()
                self.advance()

                var_names = []

                if self.current_tok.type == GL_IDENTIFIER:
                    var_names.append(self.current_tok)
                    res.register_advancement()
                    self.advance()

                    while self.current_tok.type == GL_COMMA:
                        res.register_advancement()
                        self.advance()

                        if self.current_tok.type == GL_IDENTIFIER:
                            var_names.append(self.current_tok)
                            res.register_advancement()
                            self.advance()
                        else:
                            return res.failure(
                                InvalidSyntaxError(
                                    self.current_tok.pos_start,
                                    self.current_tok.pos_end,
                                    "Expected identifier",
                                )
                            )

                if self.current_tok.type != GL_RSQUARE:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected ']'",
                        )
                    )
                res.register_advancement()
                self.advance()

                if self.current_tok.type != GL_EQ:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected '='",
                        )
                    )
                res.register_advancement()
                self.advance()

                expr = res.register(self.expr())
                if res.error:
                    return res

                return res.success(MultiVarAssignNode(var_names, expr))

            elif self.current_tok.type == GL_IDENTIFIER:
                var_name = self.current_tok
                res.register_advancement()
                self.advance()

                if self.current_tok.type == GL_LSQUARE:
                    res.register_advancement()
                    self.advance()

                    index_expr = res.register(self.expr())
                    if res.error:
                        return res

                    if self.current_tok.type != GL_RSQUARE:
                        return res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected ']'",
                            )
                        )
                    res.register_advancement()
                    self.advance()

                    if self.current_tok.type != GL_EQ:
                        return res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected '='",
                            )
                        )
                    res.register_advancement()
                    self.advance()

                    value_expr = res.register(self.expr())
                    if res.error:
                        return res

                    return res.success(
                        ListSetNode(VarAccessNode(var_name), index_expr, value_expr)
                    )

                elif self.current_tok.type == GL_EQ:
                    res.register_advancement()
                    self.advance()

                    expr = res.register(self.expr())
                    if res.error:
                        return res
                    return res.success(
                        VarAssignNode(var_name, expr, is_declaration=True)
                    )

                else:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected '=' or '['",
                        )
                    )

            else:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected identifier or '['",
                    )
                )

        if self.current_tok.matches(GL_KEYWORD, "RETURN"):
            res.register_advancement()
            self.advance()
            pos_start = self.current_tok.pos_start.copy()

            expr = res.register(self.expr())
            if res.error:
                return res

            return res.success(ReturnNode(expr, pos_start, expr.pos_end))

        if self.current_tok.matches(GL_KEYWORD, "DEF"):
            return self.fun_def()

        if self.current_tok.matches(GL_KEYWORD, "CLASS"):
            return self.class_def()

        if self.current_tok.matches(GL_KEYWORD, "ENUM"):
            return self.enum_def()

        if self.current_tok.matches(GL_KEYWORD, "TRY"):
            return self.try_expr()

        if self.current_tok.matches(GL_KEYWORD, "THROW"):
            res.register_advancement()
            self.advance()

            expr = res.register(self.expr())
            if res.error:
                return res

            return res.success(
                ThrowNode(expr, self.current_tok.pos_start, expr.pos_end)
            )

        if self.current_tok.matches(GL_KEYWORD, "SWITCH"):
            res.register_advancement()
            self.advance()
            return self.switch_expr()

        expr = res.register(self.expr())
        if res.error:
            return res
        return res.success(expr)

    def parse_iter_vars(self):
        var_toks = []
        res = ParseResult()

        if self.current_tok.type == GL_LSQUARE:
            res.register_advancement()
            self.advance()

            if self.current_tok.type == GL_IDENTIFIER:
                var_toks.append(self.current_tok)
                res.register_advancement()
                self.advance()

                while self.current_tok.type == GL_COMMA:
                    res.register_advancement()
                    self.advance()

                    if self.current_tok.type != GL_IDENTIFIER:
                        return None, res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected identifier",
                            )
                        )

                    var_toks.append(self.current_tok)
                    res.register_advancement()
                    self.advance()

            if self.current_tok.type != GL_RSQUARE:
                return None, res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected ']'",
                    )
                )
            res.register_advancement()
            self.advance()

            return var_toks, res

        elif self.current_tok.type == GL_IDENTIFIER:
            var_toks.append(self.current_tok)
            res.register_advancement()
            self.advance()
            return var_toks, res

        return None, None

    def for_expr(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "FOR"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'FOR'",
                )
            )

        res.register_advancement()
        self.advance()

        var_name_toks, var_res = self.parse_iter_vars()
        if var_res and var_res.error:
            return res.failure(var_res.error)

        if not var_name_toks:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected identifier or '[...]' for loop variable",
                )
            )
        res.advance_count += var_res.advance_count

        if not self.current_tok.matches(GL_KEYWORD, "IN"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'IN'",
                )
            )

        res.register_advancement()
        self.advance()

        iterable_node = res.register(self.expr())
        if res.error:
            return res

        self.loop_count += 1
        body_node = res.register(self.statement_list(("ENDFOR",)))
        self.loop_count -= 1
        if res.error:
            return res

        if not self.current_tok.matches(GL_KEYWORD, "ENDFOR"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDFOR'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(ForNode(var_name_toks, iterable_node, body_node))

    def while_expr(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "WHILE"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'WHILE'",
                )
            )

        res.register_advancement()
        self.advance()

        condition = res.register(self.expr())
        if res.error:
            return res

        self.loop_count += 1

        body = res.register(self.statement_list(("ENDWHILE",)))

        self.loop_count -= 1

        if res.error:
            return res

        if not self.current_tok.matches(GL_KEYWORD, "ENDWHILE"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDWHILE'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(WhileNode(condition, body))

    def logic_expr(self):
        return self.bin_op(
            self.comp_expr,
            ((GL_KEYWORD, "AND"), (GL_KEYWORD, "OR")),
        )

    def ternary_expr(self):
        res = ParseResult()

        condition_node = res.register(self.logic_expr())
        if res.error:
            return res

        if self.current_tok.type == GL_QMARK:
            res.register_advancement()
            self.advance()

            true_case_node = res.register(self.expr())
            if res.error:
                return res

            if self.current_tok.type != GL_COLON:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected ':' in ternary operator",
                    )
                )

            res.register_advancement()
            self.advance()

            false_case_node = res.register(self.ternary_expr())
            if res.error:
                return res

            return res.success(
                TernaryOpNode(condition_node, true_case_node, false_case_node)
            )

        return res.success(condition_node)

    def expr(self):
        res = ParseResult()

        node = res.register(self.ternary_expr())

        if res.error:
            return res

        if self.current_tok.type in (
            GL_EQ,
            GL_PLUSEQ,
            GL_MINUSEQ,
            GL_MULEQ,
            GL_DIVEQ,
            GL_POWEQ,
            GL_MODEQ,
            GL_FLOORDIVEQ,
            GL_BIT_ANDEQ,
            GL_BIT_OREQ,
            GL_BIT_XOREQ,
            GL_LSHIFTEQ,
            GL_RSHIFTEQ,
        ):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()

            expr = res.register(self.expr())
            if res.error:
                return res

            if op_tok.type != GL_EQ:
                bin_op_type = None
                if op_tok.type == GL_PLUSEQ:
                    bin_op_type = GL_PLUS
                elif op_tok.type == GL_MINUSEQ:
                    bin_op_type = GL_MINUS
                elif op_tok.type == GL_MULEQ:
                    bin_op_type = GL_MUL
                elif op_tok.type == GL_DIVEQ:
                    bin_op_type = GL_DIV
                elif op_tok.type == GL_POWEQ:
                    bin_op_type = GL_POW
                elif op_tok.type == GL_MODEQ:
                    bin_op_type = GL_MOD
                elif op_tok.type == GL_FLOORDIVEQ:
                    bin_op_type = GL_FLOORDIV
                elif op_tok.type == GL_BIT_ANDEQ:
                    bin_op_type = GL_BIT_AND
                elif op_tok.type == GL_BIT_OREQ:
                    bin_op_type = GL_BIT_OR
                elif op_tok.type == GL_BIT_XOREQ:
                    bin_op_type = GL_BIT_XOR
                elif op_tok.type == GL_LSHIFTEQ:
                    bin_op_type = GL_LSHIFT
                elif op_tok.type == GL_RSHIFTEQ:
                    bin_op_type = GL_RSHIFT

                expr = BinOpNode(
                    node, Token(bin_op_type, pos_start=op_tok.pos_start), expr
                )

            if isinstance(node, VarAccessNode):
                return res.success(
                    VarAssignNode(node.var_name_tok, expr, is_declaration=False)
                )
            elif isinstance(node, GetAttrNode):
                return res.success(
                    SetAttrNode(node.object_node, node.attr_name_tok, expr)
                )
            elif isinstance(node, ListAccessNode):
                return res.success(ListSetNode(node.list_node, node.index_node, expr))
            else:
                return res.failure(
                    InvalidSyntaxError(
                        node.pos_start, node.pos_end, "Invalid assignment target"
                    )
                )

        return res.success(node)

    def comp_expr(self):
        res = ParseResult()

        if self.current_tok.matches(GL_KEYWORD, "NOT"):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()

            node = res.register(self.comp_expr())
            if res.error:
                return res
            return res.success(UnaryOpNode(op_tok, node))

        node = res.register(self.bitwise_or_expr())
        if res.error:
            return res

        ops = []

        while self.current_tok.type in (GL_EE, GL_NE, GL_LT, GL_GT, GL_LTE, GL_GTE) or (
            self.current_tok.type == GL_KEYWORD
            and self.current_tok.value in ("IS", "INSTANCEOF")
        ):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()

            right_expr = res.register(self.bitwise_or_expr())
            if res.error:
                return res

            ops.append((op_tok, right_expr))

        if not ops:
            return res.success(node)

        if len(ops) == 1:
            op_tok, right_node = ops[0]
            return res.success(BinOpNode(node, op_tok, right_node))

        return res.success(ChainedCompNode(node, ops))

    def arith_expr(self):
        return self.bin_op(self.term, (GL_PLUS, GL_MINUS))

    def term(self):
        return self.bin_op(self.factor, (GL_MUL, GL_DIV, GL_MOD, GL_FLOORDIV))

    def factor(self):
        res = ParseResult()
        tok = self.current_tok

        if tok.type in (GL_PLUS, GL_MINUS):
            res.register_advancement()
            self.advance()
            factor = res.register(self.factor())
            if res.error:
                return res
            return res.success(UnaryOpNode(tok, factor))

        elif tok.type in (GL_PLUSPLUS, GL_MINUSMINUS):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()

            node = res.register(self.call())
            if res.error:
                return res

            if not isinstance(node, (VarAccessNode, GetAttrNode, ListAccessNode)):
                return res.failure(
                    InvalidSyntaxError(
                        node.pos_start,
                        op_tok.pos_end,
                        "Invalid target for pre-increment/decrement operator",
                    )
                )

            return res.success(UnaryOpNode(op_tok, node))

        elif tok.type == GL_BIT_NOT:
            res.register_advancement()
            self.advance()
            factor = res.register(self.factor())
            if res.error:
                return res
            return res.success(UnaryOpNode(tok, factor))

        return self.power()

    def power(self):
        return self.bin_op(self.call, (GL_POW,), self.factor)

    def call(self):
        res = ParseResult()
        atom = res.register(self.atom())
        if res.error:
            return res

        while True:
            if self.current_tok.type == GL_LPAREN:
                res.register_advancement()
                self.advance()
                arg_nodes = []

                if self.current_tok.type != GL_RPAREN:
                    arg_nodes.append(res.register(self.expr()))
                    if res.error:
                        return res

                    while self.current_tok.type == GL_COMMA:
                        res.register_advancement()
                        self.advance()

                        arg_nodes.append(res.register(self.expr()))
                        if res.error:
                            return res

                if self.current_tok.type != GL_RPAREN:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected ',' or ')'",
                        )
                    )

                res.register_advancement()
                self.advance()
                atom = CallNode(atom, arg_nodes)

            elif self.current_tok.type == GL_DOT:
                res.register_advancement()
                self.advance()

                if self.current_tok.type != GL_IDENTIFIER:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected identifier after '.'",
                        )
                    )

                attr_name_tok = self.current_tok
                res.register_advancement()
                self.advance()

                atom = GetAttrNode(atom, attr_name_tok)

            elif self.current_tok.type == GL_LSQUARE:
                res.register_advancement()
                self.advance()

                start_node = res.register(self.expr())
                if res.error:
                    return res

                if self.current_tok.type == GL_COLON:
                    res.register_advancement()
                    self.advance()

                    end_node = None
                    if self.current_tok.type != GL_RSQUARE:
                        end_node = res.register(self.expr())
                        if res.error:
                            return res

                    if self.current_tok.type != GL_RSQUARE:
                        return res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected ']'",
                            )
                        )

                    res.register_advancement()
                    self.advance()
                    atom = SliceAccessNode(atom, start_node, end_node)

                else:
                    if self.current_tok.type != GL_RSQUARE:
                        return res.failure(
                            InvalidSyntaxError(
                                self.current_tok.pos_start,
                                self.current_tok.pos_end,
                                "Expected ']'",
                            )
                        )

                    res.register_advancement()
                    self.advance()
                    atom = ListAccessNode(atom, start_node)

            else:
                break

        if self.current_tok.type in (GL_PLUSPLUS, GL_MINUSMINUS):
            if not isinstance(atom, (VarAccessNode, GetAttrNode, ListAccessNode)):
                return res.failure(
                    InvalidSyntaxError(
                        atom.pos_start,
                        self.current_tok.pos_end,
                        "Invalid target for post-increment/decrement operator",
                    )
                )

            op_tok = self.current_tok
            res.register_advancement()
            self.advance()
            atom = PostOpNode(atom, op_tok)

        return res.success(atom)

    def atom(self):
        res = ParseResult()
        tok = self.current_tok

        if tok.type == GL_LBRACE:
            return self.dict_expr()

        if tok.type in (GL_INT, GL_FLOAT):
            res.register_advancement()
            self.advance()
            return res.success(NumberNode(tok))

        elif tok.type == GL_STRING:
            res.register_advancement()
            self.advance()
            return res.success(StringNode(tok))

        elif tok.type == GL_IDENTIFIER:
            res.register_advancement()
            self.advance()
            return res.success(VarAccessNode(tok))

        elif tok.matches(GL_KEYWORD, "THIS"):
            res.register_advancement()
            self.advance()
            return res.success(VarAccessNode(tok))

        elif tok.matches(GL_KEYWORD, "SUPER"):
            res.register_advancement()
            self.advance()
            return res.success(VarAccessNode(tok))

        elif tok.type == GL_LPAREN:
            res.register_advancement()
            self.advance()
            expr = res.register(self.expr())
            if res.error:
                return res
            if self.current_tok.type == GL_RPAREN:
                res.register_advancement()
                self.advance()
                return res.success(expr)
            else:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected ')'",
                    )
                )

        elif tok.type == GL_LSQUARE:
            return self.list_expr()

        elif tok.matches(GL_KEYWORD, "DEF"):
            return self.fun_def()

        elif tok.matches(GL_KEYWORD, "CLASS"):
            return self.class_def()

        elif tok.matches(GL_KEYWORD, "NEW"):
            return self.new_instance()

        return res.failure(
            InvalidSyntaxError(
                tok.pos_start,
                tok.pos_end,
                "Expected int, float, string, identifier, '+', '-', '++', '--', '(', '[', 'DEF', 'CLASS', or 'NEW'",
            )
        )

    def list_expr(self):
        res = ParseResult()
        element_nodes = []
        pos_start = self.current_tok.pos_start.copy()

        if self.current_tok.type != GL_LSQUARE:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start, self.current_tok.pos_end, "Expected '['"
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type == GL_RSQUARE:
            res.register_advancement()
            self.advance()
            return res.success(
                ListNode([], pos_start, self.current_tok.pos_start.copy())
            )

        first_expr = res.register(self.expr())
        if res.error:
            return res

        if self.current_tok.matches(GL_KEYWORD, "FOR"):
            iteration_specs = []

            while self.current_tok.matches(GL_KEYWORD, "FOR"):
                res.register_advancement()
                self.advance()

                var_name_toks, var_res = self.parse_iter_vars()
                if var_res and var_res.error:
                    return res.failure(var_res.error)

                if not var_name_toks:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected identifier or '[...]'",
                        )
                    )
                res.advance_count += var_res.advance_count
                if not self.current_tok.matches(GL_KEYWORD, "IN"):
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected 'IN'",
                        )
                    )

                res.register_advancement()
                self.advance()

                iterable = res.register(self.expr())
                if res.error:
                    return res

                condition_node = None
                if self.current_tok.matches(GL_KEYWORD, "IF"):
                    res.register_advancement()
                    self.advance()

                    condition_node = res.register(self.expr())
                    if res.error:
                        return res

                iteration_specs.append((var_name_toks, iterable, condition_node))

            if self.current_tok.type != GL_RSQUARE:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected ']'",
                    )
                )

            res.register_advancement()
            self.advance()

            return res.success(
                ListCompNode(
                    first_expr,
                    iteration_specs,
                    pos_start,
                    self.current_tok.pos_start.copy(),
                )
            )

        element_nodes.append(first_expr)

        while self.current_tok.type == GL_COMMA:
            res.register_advancement()
            self.advance()

            element_nodes.append(res.register(self.expr()))
            if res.error:
                return res

        if self.current_tok.type != GL_RSQUARE:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected ',' or ']'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(
            ListNode(element_nodes, pos_start, self.current_tok.pos_start.copy())
        )

    def fun_def(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "DEF"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'DEF'",
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type == GL_IDENTIFIER:
            var_name_tok = self.current_tok
            res.register_advancement()
            self.advance()
            if self.current_tok.type != GL_LPAREN:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected '(' after function name",
                    )
                )
        else:
            var_name_tok = None
            if self.current_tok.type != GL_LPAREN:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected '('",
                    )
                )

        res.register_advancement()
        self.advance()
        arg_name_toks = []

        if self.current_tok.type != GL_RPAREN:
            if self.current_tok.type == GL_IDENTIFIER:
                arg_name_toks.append(self.current_tok)
            elif self.current_tok.matches(GL_KEYWORD, "THIS"):
                arg_name_toks.append(self.current_tok)
            else:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected identifier",
                    )
                )

            res.register_advancement()
            self.advance()

            while self.current_tok.type == GL_COMMA:
                res.register_advancement()
                self.advance()

                if self.current_tok.type == GL_IDENTIFIER:
                    arg_name_toks.append(self.current_tok)
                elif self.current_tok.matches(GL_KEYWORD, "THIS"):
                    arg_name_toks.append(self.current_tok)
                else:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected identifier",
                        )
                    )

                res.register_advancement()
                self.advance()

        if self.current_tok.type != GL_RPAREN:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected ',' or ')'",
                )
            )

        res.register_advancement()
        self.advance()

        body = res.register(self.statement_list(("ENDDEF",)))

        if res.error:
            return res

        if not self.current_tok.matches(GL_KEYWORD, "ENDDEF"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDDEF'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(FunDefNode(var_name_tok, arg_name_toks, body))

    def class_def(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "CLASS"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'CLASS'",
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type != GL_IDENTIFIER:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected class name",
                )
            )

        class_name_tok = self.current_tok
        res.register_advancement()
        self.advance()

        superclass_nodes = []
        if self.current_tok.matches(GL_KEYWORD, "INHERITS"):
            res.register_advancement()
            self.advance()

            if self.current_tok.type != GL_IDENTIFIER:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected superclass name",
                    )
                )

            superclass_nodes.append(VarAccessNode(self.current_tok))
            res.register_advancement()
            self.advance()

            while self.current_tok.type == GL_COMMA:
                res.register_advancement()
                self.advance()

                if self.current_tok.type != GL_IDENTIFIER:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected superclass name",
                        )
                    )

                superclass_nodes.append(VarAccessNode(self.current_tok))
                res.register_advancement()
                self.advance()

        method_nodes = []
        static_field_nodes = []

        while self.current_tok.type != GL_EOF and not self.current_tok.matches(
            GL_KEYWORD, "ENDCLASS"
        ):

            visibility = "PUBLIC"
            is_static = False

            if self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
                "PUBLIC",
                "PRIVATE",
                "PROTECTED",
            ):
                visibility = self.current_tok.value
                res.register_advancement()
                self.advance()

            if self.current_tok.matches(GL_KEYWORD, "STATIC"):
                is_static = True
                res.register_advancement()
                self.advance()

            if self.current_tok.type == GL_KEYWORD and self.current_tok.value in (
                "PUBLIC",
                "PRIVATE",
                "PROTECTED",
            ):
                if visibility == "PUBLIC":
                    visibility = self.current_tok.value
                    res.register_advancement()
                    self.advance()

            if self.current_tok.matches(GL_KEYWORD, "DEF"):
                method_node = res.register(self.fun_def())
                if res.error:
                    return res

                method_node.visibility = visibility
                method_node.is_static = is_static

                is_constructor = method_node.var_name_tok.value == class_name_tok.value

                if not is_static:
                    if (
                        len(method_node.arg_name_toks) == 0
                        or method_node.arg_name_toks[0].value != "THIS"
                    ):
                        method_node.arg_name_toks.insert(
                            0,
                            Token(
                                GL_KEYWORD,
                                "THIS",
                                pos_start=method_node.pos_start,
                                pos_end=method_node.pos_start,
                            ),
                        )

                if is_constructor and is_static:
                    return res.failure(
                        InvalidSyntaxError(
                            method_node.pos_start,
                            method_node.pos_end,
                            f"Constructor '{class_name_tok.value}' cannot be STATIC",
                        )
                    )

                method_nodes.append(method_node)

            elif self.current_tok.matches(
                GL_KEYWORD, "LET"
            ) or self.current_tok.matches(GL_KEYWORD, "FINAL"):
                assign_node = res.register(self.statement())
                if res.error:
                    return res

                if not isinstance(
                    assign_node,
                    (VarAssignNode, FinalVarAssignNode, VisibilityStmtNode),
                ):
                    return res.failure(
                        InvalidSyntaxError(
                            assign_node.pos_start,
                            assign_node.pos_end,
                            "Expected variable declaration inside class",
                        )
                    )

                assign_node.is_static = is_static
                assign_node.target_visibility = visibility

                static_field_nodes.append(assign_node)

            elif self.current_tok.matches(GL_KEYWORD, "ENUM"):
                enum_node = res.register(self.enum_def())
                if res.error:
                    return res

                enum_node.visibility = visibility
                enum_node.is_static = True
                static_field_nodes.append(enum_node)

            else:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected 'DEF', 'LET', 'FINAL' or 'STATIC' inside class body",
                    )
                )

        if not self.current_tok.matches(GL_KEYWORD, "ENDCLASS"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDCLASS'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(
            ClassNode(
                class_name_tok, superclass_nodes, method_nodes, static_field_nodes
            )
        )

    def enum_def(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "ENUM"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENUM'",
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type != GL_IDENTIFIER:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected enum name",
                )
            )

        enum_name_tok = self.current_tok
        pos_start = enum_name_tok.pos_start.copy()
        res.register_advancement()
        self.advance()

        cases = []
        while self.current_tok.type != GL_EOF and not self.current_tok.matches(
            GL_KEYWORD, "ENDENUM"
        ):
            if self.current_tok.type != GL_IDENTIFIER:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected identifier or 'ENDENUM'",
                    )
                )

            case_name_tok = self.current_tok
            res.register_advancement()
            self.advance()

            case_val_node = None
            if self.current_tok.type == GL_EQ:
                res.register_advancement()
                self.advance()

                case_val_node = res.register(self.expr())
                if res.error:
                    return res

            cases.append((case_name_tok, case_val_node))

            if self.current_tok.type == GL_COMMA:
                res.register_advancement()
                self.advance()

        if not self.current_tok.matches(GL_KEYWORD, "ENDENUM"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDENUM'",
                )
            )

        pos_end = self.current_tok.pos_end.copy()
        res.register_advancement()
        self.advance()

        return res.success(EnumNode(enum_name_tok, cases, pos_start, pos_end))

    def new_instance(self):
        res = ParseResult()

        if not self.current_tok.matches(GL_KEYWORD, "NEW"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'NEW'",
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type != GL_IDENTIFIER:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected class name",
                )
            )

        class_name_tok = self.current_tok
        res.register_advancement()
        self.advance()

        if self.current_tok.type != GL_LPAREN:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected '(' after class name for 'NEW'",
                )
            )

        res.register_advancement()
        self.advance()
        arg_nodes = []

        if self.current_tok.type != GL_RPAREN:
            arg_nodes.append(res.register(self.expr()))
            if res.error:
                return res

            while self.current_tok.type == GL_COMMA:
                res.register_advancement()
                self.advance()

                arg_nodes.append(res.register(self.expr()))
                if res.error:
                    return res

        if self.current_tok.type != GL_RPAREN:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected ',' or ')'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(NewInstanceNode(class_name_tok, arg_nodes))

    def bin_op(self, func_a, ops, func_b=None):
        if func_b == None:
            func_b = func_a

        res = ParseResult()
        left = res.register(func_a())
        if res.error:
            return res

        while (
            self.current_tok.type in ops
            or (self.current_tok.type, self.current_tok.value) in ops
        ):
            op_tok = self.current_tok
            res.register_advancement()
            self.advance()
            right = res.register(func_b())
            if res.error:
                return res
            left = BinOpNode(left, op_tok, right)

        return res.success(left)

    def dict_expr(self):
        res = ParseResult()
        pos_start = self.current_tok.pos_start.copy()

        if self.current_tok.type != GL_LBRACE:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start, self.current_tok.pos_end, "Expected '{'"
                )
            )

        res.register_advancement()
        self.advance()

        if self.current_tok.type == GL_RBRACE:
            res.register_advancement()
            self.advance()
            return res.success(
                DictNode([], pos_start, self.current_tok.pos_start.copy())
            )

        key = res.register(self.expr())
        if res.error:
            return res

        if self.current_tok.type != GL_COLON:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start, self.current_tok.pos_end, "Expected ':'"
                )
            )
        res.register_advancement()
        self.advance()

        value = res.register(self.expr())
        if res.error:
            return res

        if self.current_tok.matches(GL_KEYWORD, "FOR"):
            iteration_specs = []

            while self.current_tok.matches(GL_KEYWORD, "FOR"):
                res.register_advancement()
                self.advance()

                var_name_toks, var_res = self.parse_iter_vars()
                if var_res and var_res.error:
                    return res.failure(var_res.error)
                if not var_name_toks:
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected identifier or '[...]'",
                        )
                    )
                res.advance_count += var_res.advance_count

                if not self.current_tok.matches(GL_KEYWORD, "IN"):
                    return res.failure(
                        InvalidSyntaxError(
                            self.current_tok.pos_start,
                            self.current_tok.pos_end,
                            "Expected 'IN'",
                        )
                    )
                res.register_advancement()
                self.advance()

                iterable = res.register(self.expr())
                if res.error:
                    return res

                condition_node = None
                if self.current_tok.matches(GL_KEYWORD, "IF"):
                    res.register_advancement()
                    self.advance()
                    condition_node = res.register(self.expr())
                    if res.error:
                        return res

                iteration_specs.append((var_name_toks, iterable, condition_node))

            if self.current_tok.type != GL_RBRACE:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected '}'",
                    )
                )
            res.register_advancement()
            self.advance()

            return res.success(
                DictCompNode(
                    key,
                    value,
                    iteration_specs,
                    pos_start,
                    self.current_tok.pos_start.copy(),
                )
            )

        kv_pairs = []
        kv_pairs.append((key, value))

        while self.current_tok.type == GL_COMMA:
            res.register_advancement()
            self.advance()

            if self.current_tok.type == GL_RBRACE:
                break

            key = res.register(self.expr())
            if res.error:
                return res

            if self.current_tok.type != GL_COLON:
                return res.failure(
                    InvalidSyntaxError(
                        self.current_tok.pos_start,
                        self.current_tok.pos_end,
                        "Expected ':'",
                    )
                )
            res.register_advancement()
            self.advance()

            value = res.register(self.expr())
            if res.error:
                return res
            kv_pairs.append((key, value))

        if self.current_tok.type != GL_RBRACE:
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start, self.current_tok.pos_end, "Expected '}'"
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(
            DictNode(kv_pairs, pos_start, self.current_tok.pos_start.copy())
        )

    def try_expr(self):
        res = ParseResult()
        res.register_advancement()
        self.advance()

        try_body = res.register(self.statement_list(("CATCH", "FINALLY", "ENDTRY")))
        if res.error:
            return res

        catch_var = None
        catch_body = None
        finally_body = None

        if self.current_tok.matches(GL_KEYWORD, "CATCH"):
            res.register_advancement()
            self.advance()

            if self.current_tok.type == GL_IDENTIFIER:
                catch_var = self.current_tok
                res.register_advancement()
                self.advance()

            catch_body = res.register(self.statement_list(("FINALLY", "ENDTRY")))
            if res.error:
                return res

        if self.current_tok.matches(GL_KEYWORD, "FINALLY"):
            res.register_advancement()
            self.advance()

            finally_body = res.register(self.statement_list(("ENDTRY",)))
            if res.error:
                return res

        if not self.current_tok.matches(GL_KEYWORD, "ENDTRY"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDTRY'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(
            TryCatchNode(
                try_body,
                catch_var,
                catch_body,
                finally_body,
                try_body.pos_start,
                self.current_tok.pos_start.copy(),
            )
        )

    def switch_expr(self):
        res = ParseResult()
        switch_value = res.register(self.expr())
        if res.error:
            return res

        cases = []
        default_case = None

        while self.current_tok.matches(GL_KEYWORD, "CASE"):
            res.register_advancement()
            self.advance()

            case_conditions = []

            case_conditions.append(res.register(self.expr()))
            if res.error:
                return res

            while self.current_tok.type == GL_COMMA:
                res.register_advancement()
                self.advance()
                case_conditions.append(res.register(self.expr()))
                if res.error:
                    return res

            if self.current_tok.type == GL_COLON:
                res.register_advancement()
                self.advance()

            body = res.register(self.statement_list(("CASE", "DEFAULT", "ENDSWITCH")))
            if res.error:
                return res

            cases.append((case_conditions, body))

        if self.current_tok.matches(GL_KEYWORD, "DEFAULT"):
            res.register_advancement()
            self.advance()

            if self.current_tok.type == GL_COLON:
                res.register_advancement()
                self.advance()

            default_case = res.register(self.statement_list(("ENDSWITCH",)))
            if res.error:
                return res

        if not self.current_tok.matches(GL_KEYWORD, "ENDSWITCH"):
            return res.failure(
                InvalidSyntaxError(
                    self.current_tok.pos_start,
                    self.current_tok.pos_end,
                    "Expected 'ENDSWITCH'",
                )
            )

        res.register_advancement()
        self.advance()

        return res.success(SwitchNode(switch_value, cases, default_case))

    def bitwise_or_expr(self):
        return self.bin_op(self.bitwise_xor_expr, (GL_BIT_OR,))

    def bitwise_xor_expr(self):
        return self.bin_op(self.bitwise_and_expr, (GL_BIT_XOR,))

    def bitwise_and_expr(self):
        return self.bin_op(self.shift_expr, (GL_BIT_AND,))

    def shift_expr(self):
        return self.bin_op(self.arith_expr, (GL_LSHIFT, GL_RSHIFT))



class Position:
    def __init__(self, idx, ln, col, fn, ftxt):
        self.idx = idx
        self.ln = ln
        self.col = col
        self.fn = fn
        self.ftxt = ftxt

    def advance(self, current_char=None):
        self.idx += 1
        self.col += 1

        if current_char == "\n":
            self.ln += 1
            self.col = 0

        return self

    def copy(self):
        return Position(self.idx, self.ln, self.col, self.fn, self.ftxt)


class Error:
    def __init__(self, pos_start, pos_end, error_name, details):
        self.pos_start = pos_start
        self.pos_end = pos_end
        self.error_name = error_name
        self.details = details

    def as_string(self):
        result = f"{self.error_name}: {self.details}\n"
        result += f"File {self.pos_start.fn}, line {self.pos_start.ln + 1}"
        return result


class IllegalCharError(Error):
    def __init__(self, pos_start, pos_end, details):
        super().__init__(pos_start, pos_end, "Illegal Character", details)


class InvalidSyntaxError(Error):
    def __init__(self, pos_start, pos_end, details=""):
        super().__init__(pos_start, pos_end, "Invalid Syntax", details)


class RTError(Error):
    def __init__(self, pos_start, pos_end, details, context, thrown_value=None):
        super().__init__(pos_start, pos_end, "Runtime Error", details)
        self.context = context
        self.thrown_value = thrown_value

    def as_string(self):
        result = self.generate_traceback()
        result += f"{self.error_name}: {self.details}"
        return result

    def generate_traceback(self):
        result = ""
        pos = self.pos_start
        ctx = self.context

        while ctx:
            result = (
                f"  File {pos.fn}, line {str(pos.ln + 1)}, in {ctx.display_name}\n"
                + result
            )
            pos = ctx.parent_entry_pos
            ctx = ctx.parent

        return "Traceback (most recent call last):\n" + result



import sys
from .errors import RTError
from .runtime import SymbolTable, Context, RTResult
from .constants import GL_IDENTIFIER
from .lexer import Token
from .nodes import *


class Value:
    def __init__(self):
        self.set_pos()
        self.set_context()

    def set_pos(self, pos_start=None, pos_end=None):
        self.pos_start = pos_start
        self.pos_end = pos_end
        return self

    def set_context(self, context=None):
        self.context = context
        return self

    def added_to(self, other):
        return None, self.illegal_operation(other)

    def subbed_by(self, other):
        return None, self.illegal_operation(other)

    def multed_by(self, other):
        return None, self.illegal_operation(other)

    def dived_by(self, other):
        return None, self.illegal_operation(other)

    def modded_by(self, other):
        return None, self.illegal_operation(other)

    def floordived_by(self, other):
        return None, self.illegal_operation(other)

    def powed_by(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_eq(self, other, visited=None):
        return None, self.illegal_operation(other)

    def get_comparison_ne(self, other):
        if not isinstance(other, List):
            return None, Value.illegal_operation(self, other)

        result, error = self.get_comparison_eq(other)
        if error:
            return None, error

        if result.is_true():
            return Number(0).set_context(self.context), None
        else:
            return Number(1).set_context(self.context), None

    def get_comparison_lt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gt(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_lte(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_gte(self, other):
        return None, self.illegal_operation(other)

    def get_comparison_is(self, other):
        return Number(1 if self is other else 0).set_context(self.context), None

    def get_comparison_instanceof(self, other):
        if isinstance(other, Type):
            if other.name == "Number" and isinstance(self, Number):
                return Number.true, None
            if other.name == "String" and isinstance(self, String):
                return Number.true, None
            if other.name == "List" and isinstance(self, List):
                return Number.true, None
            if other.name == "Dict" and isinstance(self, Dict):
                return Number.true, None
            if other.name == "Function" and isinstance(self, BaseFunction):
                return Number.true, None
            if other.name == "Object":
                return Number.true, None
            return Number.false, None

        if isinstance(other, Class):
            return Number.false, None

        return None, RTError(
            self.pos_start,
            self.pos_end,
            "Right operand of INSTANCEOF must be a Class or Type",
            self.context,
        )

    def anded_by(self, other):
        is_true = self.is_true() and other.is_true()
        return Number(1 if is_true else 0).set_context(self.context), None

    def ored_by(self, other):
        is_true = self.is_true() or other.is_true()
        return Number(1 if is_true else 0).set_context(self.context), None

    def notted(self):
        return None, self.illegal_operation()

    def execute(self, args, interpreter=None):
        return RTResult().failure(self.illegal_operation())

    def get_attr(self, name_tok):
        return None, self.illegal_operation()

    def set_attr(self, name_tok, value, context=None, visibility=None, as_final=False):
        return None, self.illegal_operation()

    def get_element_at(self, index):
        return None, self.illegal_operation()

    def set_element_at(self, index, value):
        return None, self.illegal_operation()

    def is_true(self):
        return True

    def copy(self):
        raise Exception("No copy method defined")

    def bitted_and_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_or_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_xor_by(self, other):
        return None, self.illegal_operation(other)

    def lshifted_by(self, other):
        return None, self.illegal_operation(other)

    def rshifted_by(self, other):
        return None, self.illegal_operation(other)

    def bitted_not(self):
        return None, self.illegal_operation()

    def illegal_operation(self, other=None):
        if not other:
            other = self
        return RTError(self.pos_start, other.pos_end, "Illegal operation", self.context)


class Number(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, Number):
            return Number(self.value + other.value).set_context(self.context), None
        elif isinstance(other, String):
            return String(str(self.value) + other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def subbed_by(self, other):
        if isinstance(other, Number):
            return Number(self.value - other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def multed_by(self, other):
        if isinstance(other, Number):
            return Number(self.value * other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def dived_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value / other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def modded_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value % other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def floordived_by(self, other):
        if isinstance(other, Number):
            if other.value == 0:
                return None, RTError(
                    other.pos_start, other.pos_end, "Division by zero", self.context
                )
            return Number(self.value // other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_is(self, other):
        return Number(1 if self is other else 0).set_context(self.context), None

    def powed_by(self, other):
        if isinstance(other, Number):
            if other.value > 10000:
                return None, RTError(
                    other.pos_start,
                    other.pos_end,
                    "Exponent too large (limit: 10000)",
                    self.context,
                )
            return Number(self.value**other.value).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_eq(self, other, visited=None):
        if isinstance(other, Number):
            return (
                Number(int(self.value == other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value != other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, Number):
            return Number(int(self.value < other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, Number):
            return Number(int(self.value > other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value <= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value >= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def anded_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value and other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def ored_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value or other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def notted(self):
        return Number(1 if self.value == 0 else 0).set_context(self.context), None

    def bitted_and_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) & int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_or_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) | int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_xor_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) ^ int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def lshifted_by(self, other):
        if isinstance(other, Number):
            try:
                result = int(self.value) << int(other.value)
                return Number(result).set_context(self.context), None
            except ValueError:
                return None, RTError(
                    other.pos_start, other.pos_end, "Negative shift count", self.context
                )
        return None, Value.illegal_operation(self, other)

    def rshifted_by(self, other):
        if isinstance(other, Number):
            return (
                Number(int(self.value) >> int(other.value)).set_context(self.context),
                None,
            )
        return None, Value.illegal_operation(self, other)

    def bitted_not(self):
        return Number(~int(self.value)).set_context(self.context), None

    def is_true(self):
        return self.value != 0

    def copy(self):
        copy = Number(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return str(self.value)


Number.false = Number(0)
Number.true = Number(1)
Number.null = Number(0)


class String(Value):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def added_to(self, other):
        if isinstance(other, String):
            return String(self.value + other.value).set_context(self.context), None
        elif isinstance(other, Number):
            return String(self.value + str(other.value)).set_context(self.context), None
        else:
            return None, self.illegal_operation(other)

    def get_comparison_eq(self, other, visited=None):
        if isinstance(other, String):
            return (
                Number(int(self.value == other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_ne(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value != other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lt(self, other):
        if isinstance(other, String):
            return Number(int(self.value < other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gt(self, other):
        if isinstance(other, String):
            return Number(int(self.value > other.value)).set_context(self.context), None
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_lte(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value <= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_comparison_gte(self, other):
        if isinstance(other, String):
            return (
                Number(int(self.value >= other.value)).set_context(self.context),
                None,
            )
        else:
            return None, Value.illegal_operation(self, other)

    def get_element_at(self, index):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "String index must be a Number",
                self.context,
            )

        try:
            val = self.value[int(index.value)]
            return String(val).set_context(self.context), None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"String index {index.value} out of bounds",
                self.context,
            )

    def is_true(self):
        return len(self.value) > 0

    def copy(self):
        copy = String(self.value)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return self.value


class List(Value):
    def __init__(self, elements):
        super().__init__()
        self.elements = elements

    def is_true(self):
        return len(self.elements) > 0

    def added_to(self, other):
        if isinstance(other, List):
            new_list = List(self.elements + other.elements)
            new_list.set_context(self.context)
            return new_list, None
        else:
            return None, self.illegal_operation(other)

    def multed_by(self, other):
        if isinstance(other, Number):
            new_list = List(self.elements * int(other.value))
            new_list.set_context(self.context)
            return new_list, None
        return None, self.illegal_operation(other)

    def get_element_at(self, index):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "List index must be a Number",
                self.context,
            )

        try:
            element = self.elements[int(index.value)]
            return element, None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"List index {index.value} out of bounds",
                self.context,
            )

    def set_element_at(self, index, value):
        if not isinstance(index, Number):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "List index must be a Number",
                self.context,
            )

        try:
            self.elements[int(index.value)] = value
            return value, None
        except IndexError:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"List index {index.value} out of bounds",
                self.context,
            )

    def get_comparison_eq(self, other, visited=None):
        if not isinstance(other, List):
            return None, Value.illegal_operation(self, other)

        if len(self.elements) != len(other.elements):
            return Number(0).set_context(self.context), None

        if visited is None:
            visited = set()

        pair = (id(self), id(other))
        if pair in visited:
            return Number(1).set_context(self.context), None

        visited.add(pair)

        try:
            for i in range(len(self.elements)):
                result, error = self.elements[i].get_comparison_eq(
                    other.elements[i], visited
                )
                if error:
                    return None, error
                if not result.is_true():
                    return Number(0).set_context(self.context), None
        finally:
            visited.remove(pair)

        return Number(1).set_context(self.context), None

    def get_comparison_ne(self, other):
        if not isinstance(other, List):
            return None, Value.illegal_operation(self, other)

        result, error = self.get_comparison_eq(other)
        if error:
            return None, error

        if result.is_true():
            return Number(0).set_context(self.context), None
        else:
            return Number(1).set_context(self.context), None

    def copy(self):
        copy = List(self.elements[:])
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return self.to_string([])

    def to_string(self, visited):
        if self in visited:
            return "[...]"
        visited.append(self)
        s = f'[{", ".join([x.to_string(visited) if isinstance(x, (List, Dict)) else repr(x) for x in self.elements])}]'
        visited.pop()
        return s


class Dict(Value):
    def __init__(self, elements):
        super().__init__()
        self.elements = elements

    def is_true(self):
        return len(self.elements) > 0

    def copy(self):
        copy = Dict(self.elements.copy())
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def added_to(self, other):
        if isinstance(other, Dict):
            new_dict = self.copy()
            new_dict.elements.update(other.elements)
            return new_dict, None
        return None, self.illegal_operation(other)

    def get_element_at(self, key):
        if not isinstance(key, (Number, String)):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "Key must be a Number or String",
                self.context,
            )

        val = self.elements.get(key.value)
        if val is None:
            return None, RTError(
                self.pos_start,
                self.pos_end,
                f"Key '{key.value}' not found",
                self.context,
            )
        return val, None

    def set_element_at(self, key, value):
        if not isinstance(key, (Number, String)):
            return None, RTError(
                self.pos_start,
                self.pos_end,
                "Key must be a Number or String",
                self.context,
            )

        self.elements[key.value] = value
        return value, None

    def get_comparison_eq(self, other, visited=None):
        if not isinstance(other, Dict):
            return None, Value.illegal_operation(self, other)

        if len(self.elements) != len(other.elements):
            return Number(0).set_context(self.context), None

        if visited is None:
            visited = set()

        pair = (id(self), id(other))
        if pair in visited:
            return Number(1).set_context(self.context), None

        visited.add(pair)

        try:
            for key, value in self.elements.items():
                if key not in other.elements:
                    return Number(0).set_context(self.context), None

                other_val = other.elements[key]
                result, error = value.get_comparison_eq(other_val, visited)
                if error:
                    return None, error

                if not result.is_true():
                    return Number(0).set_context(self.context), None
        finally:
            visited.remove(pair)

        return Number(1).set_context(self.context), None

    def get_comparison_ne(self, other):
        if not isinstance(other, Dict):
            return None, Value.illegal_operation(self, other)

        result, error = self.get_comparison_eq(other)
        if error:
            return None, error

        if result.is_true():
            return Number(0).set_context(self.context), None
        else:
            return Number(1).set_context(self.context), None

    def __repr__(self):
        return self.to_string([])

    def to_string(self, visited):
        if self in visited:
            return "{...}"
        visited.append(self)
        kv_strings = []
        for key, value in self.elements.items():
            val_str = (
                value.to_string(visited)
                if isinstance(value, (List, Dict))
                else repr(value)
            )
            kv_strings.append(f"{repr(key)}: {val_str}")
        s = f"{{{', '.join(kv_strings)}}}"
        visited.pop()
        return s


class BaseFunction(Value):
    def __init__(self, name):
        super().__init__()
        self.name = name or "<anonymous>"

    def generate_new_context(self):
        new_context = Context(self.name, self.context, self.pos_start)
        new_context.symbol_table = SymbolTable(self.context.symbol_table)
        return new_context

    def get_comparison_eq(self, other, visited=None):
        if isinstance(other, BaseFunction):
            return Number(1 if self is other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def get_comparison_ne(self, other):
        if isinstance(other, BaseFunction):
            return Number(1 if self is not other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def check_args(self, arg_names, args):
        res = RTResult()
        if len(args) != len(arg_names):
            return res.failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    f"Incorrect argument count for '{self.name}'. Expected {len(arg_names)}, got {len(args)}",
                    self.context,
                )
            )
        return res.success(None)

    def populate_args(self, arg_names, args, new_context):
        for i in range(len(args)):
            arg_name = arg_names[i]
            arg_value = args[i]

            if isinstance(arg_value, BaseFunction):
                pass
            else:
                pass

            new_context.symbol_table.set(arg_name, arg_value)

    def check_and_populate_args(self, arg_names, args, new_context):
        res = RTResult()
        res.register(self.check_args(arg_names, args))
        if res.error:
            return res
        self.populate_args(arg_names, args, new_context)
        return res.success(None)

    def execute(self, args, interpreter):
        return RTResult().failure(
            RTError(
                self.pos_start,
                self.pos_end,
                "BaseFunction cannot be executed",
                self.context,
            )
        )

    def copy(self):
        raise Exception("Cannot copy a BaseFunction")

    def __repr__(self):
        return f"<function {self.name}>"


class Function(BaseFunction):
    def __init__(
        self,
        name,
        body_node,
        arg_name_toks,
        parent_context,
        visibility="PUBLIC",
        defining_class=None,
        is_static=False,
    ):
        super().__init__(name)
        self.body_node = body_node
        self.arg_name_toks = arg_name_toks
        self.arg_names = [tok.value for tok in arg_name_toks]
        self.context = parent_context
        self.visibility = visibility
        self.defining_class = defining_class
        self.is_static = is_static

    def execute(self, args, interpreter):
        res = RTResult()

        new_context = self.generate_new_context()

        new_context.active_class = self.defining_class

        if new_context.depth > 250:
            return res.failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    "Recursion limit exceeded",
                    self.context,
                )
            )

        res.register(self.check_and_populate_args(self.arg_names, args, new_context))
        if res.error:
            return res

        value_result = interpreter.visit(self.body_node, new_context)

        if value_result.error:
            return value_result

        if value_result.should_return:
            return res.success(value_result.return_value)

        return res.success(value_result.value or Number.null)

    def copy(self):
        copy = Function(
            self.name,
            self.body_node,
            self.arg_name_toks,
            self.context,
            self.visibility,
            self.defining_class,
            self.is_static,
        )

        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy


class FunctionGroup(BaseFunction):
    def __init__(self, name):
        super().__init__(name)
        self.functions = {}
        self.visibility = "PUBLIC"
        self.is_static = False
        self.defining_class = None

    def add_function(self, func):
        arity = len(func.arg_names)

        if arity in self.functions:
            raise Exception(
                f"Duplicate function definition: '{self.name}' with {arity} arguments is already defined."
            )

        self.functions[arity] = func

        if len(self.functions) == 1:
            self.visibility = getattr(func, "visibility", "PUBLIC")
            self.is_static = getattr(func, "is_static", False)
            self.defining_class = getattr(func, "defining_class", None)

    def execute(self, args, interpreter):
        arity = len(args)
        if arity not in self.functions:
            return RTResult().failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    f"No variant of function '{self.name}' found that accepts {arity} arguments",
                    self.context,
                )
            )

        return self.functions[arity].execute(args, interpreter)

    def bind_to_instance(self, instance):
        return BoundMethod(self.name, self, instance)

    def copy(self):
        copy = FunctionGroup(self.name)
        copy.visibility = self.visibility
        copy.is_static = self.is_static
        copy.defining_class = self.defining_class

        for arity, func in self.functions.items():
            copy.functions[arity] = func.copy()
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return f"<function group {self.name}>"


class Class(BaseFunction):
    def __init__(self, name, superclasses, methods, static_symbol_table=None, mro=None):
        super().__init__(name)
        self.superclasses = superclasses
        self.methods = methods
        self.static_symbol_table = (
            static_symbol_table if static_symbol_table else SymbolTable()
        )
        self.mro = mro if mro else [self]

    def instantiate(self, args, context=None, interpreter=None):
        res = RTResult()
        instance = Instance(self)

        constructor_name = None

        for cls in self.mro:
            if cls.name in cls.methods:
                constructor_name = cls.name
                break

        if constructor_name:
            fake_tok = Token(
                GL_IDENTIFIER, constructor_name, self.pos_start, self.pos_end
            )
            init_method, error = self.get_attr(fake_tok, context, allow_instance=True)

            if error:
                return res.failure(error)

            bound_init = init_method.copy().bind_to_instance(instance)
            res.register(bound_init.execute(args, interpreter))
            if res.error:
                return res
        else:
            if len(args) > 0:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"'{self.name}' does not have a constructor that accepts arguments",
                        self.context,
                    )
                )

        return res.success(instance)

    def execute(self, args, interpreter=None):
        return RTResult().failure(
            RTError(
                self.pos_start,
                self.pos_end,
                f"Class '{self.name}' must be instantiated using 'NEW'",
                self.context,
            )
        )

    def set_attr(self, name_tok, value, context=None, visibility=None, as_final=False):
        name = name_tok.value

        if visibility == "FINAL":
            visibility = "PUBLIC"
            as_final = True

        if as_final or visibility is not None:
            if name in self.static_symbol_table.symbols:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Static field '{name}' is already defined",
                    context,
                )
            self.static_symbol_table.set(
                name, value, visibility=(visibility or "PUBLIC"), as_final=as_final
            )
            return value, None

        if self.static_symbol_table.get(name):
            if name in self.static_symbol_table.finals:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot reassign static constant '{name}'",
                    context,
                )
            err = self.static_symbol_table.update(name, value)
            if err:
                return None, RTError(name_tok.pos_start, name_tok.pos_end, err, context)
            return value, None

        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Class '{self.name}' has no static field '{name}'",
            self.context,
        )

    def get_attr(self, name_tok, context=None, allow_instance=False):
        method_name = name_tok.value

        for cls in self.mro:
            val = cls.static_symbol_table.get(method_name)
            if val:
                visibility = cls.static_symbol_table.get_visibility(method_name)
                defining_class = cls

                if visibility == "PRIVATE":
                    if not context or context.active_class != defining_class:
                        return None, RTError(
                            name_tok.pos_start,
                            name_tok.pos_end,
                            f"Cannot access private static field '{method_name}'",
                            context,
                        )

                if visibility == "PROTECTED":
                    allowed = False
                    if context and context.active_class:
                        if (
                            defining_class in context.active_class.mro
                            or context.active_class in defining_class.mro
                        ):
                            allowed = True

                    if not allowed:
                        return None, RTError(
                            name_tok.pos_start,
                            name_tok.pos_end,
                            f"Cannot access protected static field '{method_name}'",
                            context,
                        )

                return val, None

            method = cls.methods.get(method_name)
            if method:
                visibility = method.visibility
                defining_class = method.defining_class

                if visibility == "PRIVATE":
                    if not context or context.active_class != defining_class:
                        return None, RTError(
                            name_tok.pos_start,
                            name_tok.pos_end,
                            f"Cannot access private method '{method_name}' via Class",
                            context,
                        )

                if visibility == "PROTECTED":
                    allowed = False
                    if context and context.active_class:
                        if defining_class in context.active_class.mro:
                            allowed = True

                    if not allowed:
                        return None, RTError(
                            name_tok.pos_start,
                            name_tok.pos_end,
                            f"Cannot access protected method '{method_name}' via Class",
                            context,
                        )

                if method.is_static:
                    return (
                        method.copy()
                        .set_context(self.context)
                        .set_pos(name_tok.pos_start, name_tok.pos_end),
                        None,
                    )

                if context:
                    instance = context.symbol_table.get("THIS")
                    if (
                        instance
                        and isinstance(instance, Instance)
                        and self in instance.class_ref.mro
                    ):
                        return method.copy().bind_to_instance(instance), None

                return method, None

        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Class '{self.name}' has no member '{method_name}'",
            self.context,
        )

    def copy(self):
        copy = Class(
            self.name,
            self.superclasses,
            self.methods,
            self.static_symbol_table,
            self.mro,
        )
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<class {self.name}>"


class Instance(Value):
    def __init__(self, class_ref):
        super().__init__()
        self.class_ref = class_ref
        self.symbol_table = SymbolTable()

    def check_access(self, name_tok, visibility, defining_class, context):
        if visibility == "PUBLIC":
            return None

        if visibility == "PRIVATE":
            if not context or context.active_class != defining_class:
                return RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot access private member '{name_tok.value}'",
                    context,
                )

        if visibility == "PROTECTED":
            allowed = False
            if context and context.active_class:
                if (
                    defining_class in context.active_class.mro
                    or context.active_class in defining_class.mro
                ):
                    allowed = True
            if not allowed:
                return RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot access protected member '{name_tok.value}'",
                    context,
                )
        return None

    def get_attr(self, name_tok, context=None):
        name = name_tok.value

        if context and context.active_class:
            mangled_name = f"_{context.active_class.name}__{name}"
            val = self.symbol_table.get(mangled_name)
            if val:
                return val, None

        value = self.symbol_table.get(name)
        if value:
            visibility = self.symbol_table.get_visibility(name)
            error = self.check_access(name_tok, visibility, self.class_ref, context)
            if error:
                return None, error
            return value, None

        member, error = self.class_ref.get_attr(name_tok, context, allow_instance=True)
        if error:
            return None, error

        if isinstance(member, BaseFunction):
            if isinstance(member, Function) and member.is_static:
                return member, None

            error = self.check_access(
                name_tok, member.visibility, member.defining_class, context
            )
            if error:
                return None, error

            bound_method = member.copy().bind_to_instance(self)
            return bound_method, None

        return member, None

    def set_attr(self, name_tok, value, context=None, visibility=None, as_final=False):
        name = name_tok.value

        if visibility == "FINAL":
            visibility = "PUBLIC"
            as_final = True

        mangled_name = None
        if context and context.active_class:
            mangled_name = f"_{context.active_class.name}__{name}"

        if visibility is not None or as_final:
            target_name = mangled_name if visibility == "PRIVATE" else name

            if target_name in self.symbol_table.symbols:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Attribute '{name}' is already defined",
                    context,
                )

            self.symbol_table.set(
                target_name,
                value,
                visibility=(visibility or "PUBLIC"),
                as_final=as_final,
            )

            if visibility == "PRIVATE" and self.symbol_table.get(name):
                if name in self.symbol_table.finals:
                    return None, RTError(
                        name_tok.pos_start,
                        name_tok.pos_end,
                        f"Cannot shadow constant '{name}' with a private variable",
                        context,
                    )
                self.symbol_table.remove(name)

            return value, None

        if mangled_name and self.symbol_table.get(mangled_name):
            if mangled_name in self.symbol_table.finals:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot reassign constant '{name}'",
                    context,
                )
            self.symbol_table.set(mangled_name, value, visibility="PRIVATE")
            return value, None

        if self.symbol_table.get(name):
            current_vis = self.symbol_table.get_visibility(name)
            error = self.check_access(name_tok, current_vis, self.class_ref, context)
            if error:
                return None, error

            if name in self.symbol_table.finals:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot reassign constant '{name}'",
                    context,
                )

            self.symbol_table.set(name, value, visibility=current_vis)
            return value, None

        if self.class_ref.static_symbol_table.get(name):
            if name in self.class_ref.static_symbol_table.finals:
                return None, RTError(
                    name_tok.pos_start,
                    name_tok.pos_end,
                    f"Cannot shadow static constant '{name}' with an instance variable",
                    context,
                )

        self.symbol_table.set(name, value, visibility="PUBLIC")
        return value, None

    def get_comparison_eq(self, other, visited=None):
        if isinstance(other, Instance):
            return Number(1 if self is other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def get_comparison_ne(self, other):
        if isinstance(other, Instance):
            return Number(1 if self is not other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def get_comparison_instanceof(self, other):
        if isinstance(other, Class):
            is_instance = other in self.class_ref.mro
            return Number(1 if is_instance else 0).set_context(self.context), None

        if isinstance(other, Type):
            if other.name == "Object":
                return Number.true, None
            return Number.false, None

        return None, RTError(
            self.pos_start,
            self.pos_end,
            "Right operand of INSTANCEOF must be a Class or Type",
            self.context,
        )

    def copy(self):
        copy = Instance(self.class_ref)
        copy.symbol_table = self.symbol_table.copy()
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<{self.class_ref.name} instance>"


class Super(Value):
    def __init__(self, instance, start_class):
        super().__init__()
        self.instance = instance
        self.start_class = start_class

    def get_attr(self, name_tok, context=None):
        method_name = name_tok.value
        mro = self.instance.class_ref.mro

        try:
            start_index = mro.index(self.start_class) + 1
        except ValueError:
            return None, RTError(
                name_tok.pos_start,
                name_tok.pos_end,
                "Current class not found in MRO",
                context,
            )

        for i in range(start_index, len(mro)):
            cls = mro[i]
            method = cls.methods.get(method_name)
            if method:
                return method.copy().bind_to_instance(self.instance), None

        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Method '{method_name}' not found in superclasses",
            context,
        )

    def execute(self, args, interpreter):
        res = RTResult()
        mro = self.instance.class_ref.mro

        try:
            start_index = mro.index(self.start_class) + 1
        except ValueError:
            return res.failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    "Current class not found in MRO",
                    self.context,
                )
            )

        method = None
        for i in range(start_index, len(mro)):
            cls = mro[i]
            if cls.name in cls.methods:
                method = cls.methods[cls.name]
                break

        if method:
            return (
                method.copy().bind_to_instance(self.instance).execute(args, interpreter)
            )

        if len(args) > 0:
            return res.failure(
                RTError(
                    self.pos_start,
                    self.pos_end,
                    "No constructor found in superclasses",
                    self.context,
                )
            )

        return res.success(Number.null)

    def copy(self):
        return (
            Super(self.instance, self.start_class)
            .set_context(self.context)
            .set_pos(self.pos_start, self.pos_end)
        )


class Type(Value):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def copy(self):
        return (
            Type(self.name)
            .set_context(self.context)
            .set_pos(self.pos_start, self.pos_end)
        )

    def __repr__(self):
        return f"<type {self.name}>"


class Enum(Value):
    def __init__(self, name, elements_dict):
        super().__init__()
        self.name = name
        self.elements_dict = elements_dict

    def get_attr(self, name_tok, context=None):
        name = name_tok.value
        if name in self.elements_dict:
            val = self.elements_dict[name]
            return val.copy().set_pos(name_tok.pos_start, name_tok.pos_end), None
        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Enum '{self.name}' has no case '{name}'",
            self.context,
        )

    def set_attr(self, name_tok, value, context=None, visibility=None, as_final=False):
        return None, RTError(
            name_tok.pos_start,
            name_tok.pos_end,
            f"Cannot reassign enum case '{name_tok.value}'",
            self.context,
        )

    def get_comparison_eq(self, other, visited=None):
        if isinstance(other, Enum):
            return Number(1 if self is other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def get_comparison_ne(self, other):
        if isinstance(other, Enum):
            return Number(1 if self is not other else 0).set_context(self.context), None
        return None, self.illegal_operation(other)

    def copy(self):
        copy = Enum(self.name, self.elements_dict)
        copy.set_pos(self.pos_start, self.pos_end)
        copy.set_context(self.context)
        return copy

    def __repr__(self):
        return f"<enum {self.name}>"


class BoundMethod(BaseFunction):
    def __init__(self, name, function_to_bind, instance):
        super().__init__(name)
        self.function_to_bind = function_to_bind
        self.instance = instance
        self.context = function_to_bind.context
        self.set_pos(function_to_bind.pos_start, function_to_bind.pos_end)

        self.visibility = getattr(function_to_bind, "visibility", "PUBLIC")
        self.defining_class = getattr(function_to_bind, "defining_class", None)
        self.is_static = getattr(function_to_bind, "is_static", False)

    def bind_to_instance(self, instance):
        return BoundMethod(self.name, self.function_to_bind, instance)

    def execute(self, args, interpreter):
        res = RTResult()

        if isinstance(self.function_to_bind, FunctionGroup):
            full_args = [self.instance] + args
            return self.function_to_bind.execute(full_args, interpreter)

        new_context = self.function_to_bind.generate_new_context()

        new_context.active_class = self.function_to_bind.defining_class

        original_arg_names = self.function_to_bind.arg_names

        new_context.symbol_table.set("THIS", self.instance)

        actual_args = args
        if len(args) > 0 and args[0] is self.instance:
            actual_args = args[1:]

        if len(original_arg_names) > 0 and original_arg_names[0] == "THIS":
            expected_arg_names = original_arg_names[1:]
        else:
            expected_arg_names = original_arg_names

        res.register(
            self.function_to_bind.check_and_populate_args(
                expected_arg_names, actual_args, new_context
            )
        )

        if res.error:
            return res

        value_result = interpreter.visit(self.function_to_bind.body_node, new_context)

        if value_result.error:
            return value_result

        if value_result.should_return:
            return res.success(value_result.return_value)

        return res.success(value_result.value or Number.null)

    def copy(self):
        return (
            BoundMethod(self.name, self.function_to_bind.copy(), self.instance)
            .set_context(self.context)
            .set_pos(self.pos_start, self.pos_end)
        )


class BuiltInFunction(BaseFunction):
    def __init__(self, name):
        super().__init__(name)

    def execute(self, args, interpreter):
        res = RTResult()

        if self.name == "INPUT":
            if len(args) > 1:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        "INPUT takes at most 1 argument",
                        self.context,
                    )
                )

            prompt = ""
            if len(args) == 1:
                prompt = str(args[0].value)

            if prompt:
                sys.stdout.write(prompt)
                sys.stdout.flush()

            text = sys.stdin.readline(4096)

            if text:
                text = text.rstrip("\n")

            return res.success(String(text))

        elif self.name == "STR":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            val = args[0]
            if isinstance(val, String):
                return res.success(String(val.value))
            else:
                return res.success(String(str(val)))

        elif self.name == "INT":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res
            arg = args[0]

            if not isinstance(arg, (Number, String)):
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Argument for INT must be a Number or String, got {type(arg).__name__}",
                        self.context,
                    )
                )

            try:
                val = int(float(arg.value))
                return res.success(Number(val))
            except ValueError:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Cannot convert '{arg.value}' to INT",
                        self.context,
                    )
                )

        elif self.name == "FLOAT":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res
            arg = args[0]

            if not isinstance(arg, (Number, String)):
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Argument for FLOAT must be a Number or String, got {type(arg).__name__}",
                        self.context,
                    )
                )

            try:
                val = float(arg.value)
                return res.success(Number(val))
            except ValueError:
                return res.failure(
                    RTError(
                        self.pos_start,
                        self.pos_end,
                        f"Cannot convert '{arg.value}' to FLOAT",
                        self.context,
                    )
                )

        elif self.name == "BOOL":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            is_true = args[0].is_true()
            return res.success(Number.true if is_true else Number.false)

        elif self.name == "LEN":
            res.register(self.check_args(["value"], args))
            if res.error:
                return res

            arg = args[0]

            if isinstance(arg, String):
                return res.success(Number(len(arg.value)))
            elif isinstance(arg, List):
                return res.success(Number(len(arg.elements)))
            elif isinstance(arg, Dict):
                return res.success(Number(len(arg.elements)))
            elif isinstance(arg, Number):
                return res.success(Number(len(str(arg.value))))
            elif isinstance(arg, (Function, BuiltInFunction, Class)):
                return res.success(Number(1))

            return res.success(Number(0))

        return res.failure(
            RTError(
                self.pos_start,
                self.pos_end,
                f"Built-in function '{self.name}' is not defined.",
                self.context,
            )
        )

    def copy(self):
        copy = BuiltInFunction(self.name)
        copy.set_context(self.context)
        copy.set_pos(self.pos_start, self.pos_end)
        return copy

    def __repr__(self):
        return f"<built-in function {self.name}>"


def bind_to_instance(self, instance):
    return BoundMethod(self.name, self, instance)


Function.bind_to_instance = bind_to_instance



import sys
import re
import io
from pathlib import Path

try:
    import resource
except ImportError:
    resource = None

if hasattr(sys, "set_int_max_str_digits"):
    sys.set_int_max_str_digits(50000)

if hasattr(sys, "setrecursionlimit"):
    sys.setrecursionlimit(2000)

from .runtime import SymbolTable, Context
from .values import Number, BuiltInFunction, List, Type
from .lexer import Lexer
from .parser import Parser
from .interpreter import Interpreter


def get_fresh_global_scope():
    scope = SymbolTable()

    scope.set("NULL", Number(0))
    scope.set("FALSE", Number(0))
    scope.set("TRUE", Number(1))

    scope.set("Number", Type("Number"))
    scope.set("String", Type("String"))
    scope.set("List", Type("List"))
    scope.set("Dict", Type("Dict"))
    scope.set("Enum", Type("Enum"))
    scope.set("Function", Type("Function"))
    scope.set("Object", Type("Object"))

    scope.set("INPUT", BuiltInFunction("INPUT"))
    scope.set("STR", BuiltInFunction("STR"))
    scope.set("INT", BuiltInFunction("INT"))
    scope.set("FLOAT", BuiltInFunction("FLOAT"))
    scope.set("BOOL", BuiltInFunction("BOOL"))

    scope.set("LEN", BuiltInFunction("LEN"))
    scope.set("LENGTH", BuiltInFunction("LEN"))

    return scope


def set_memory_limit(max_mb):
    if resource:
        try:
            soft, hard = resource.getrlimit(resource.RLIMIT_AS)
            limit_bytes = max_mb * 1024 * 1024
            resource.setrlimit(resource.RLIMIT_AS, (limit_bytes, hard))
        except Exception as e:
            sys.stderr.write(f"Warning: Could not set memory limit: {e}\n")


def run(fn, text, context=None, instruction_limit=None):
    lexer = Lexer(fn, text)
    tokens, error = lexer.make_tokens()
    if error:
        return None, error

    parser = Parser(tokens)
    ast = parser.parse()
    if ast.error:
        return None, ast.error

    interpreter = Interpreter(instruction_limit=instruction_limit)

    if context is None:
        context = Context("<program>")
        context.symbol_table = get_fresh_global_scope()

    result = interpreter.visit(ast.node, context)

    if result.should_return:
        return result.return_value, result.error

    return result.value, result.error


def is_complete(text):
    text = re.sub(r"#.*", "", text)

    temp_text = re.sub(r'""".*?"""', "", text, flags=re.DOTALL)
    if temp_text.count('"""') % 2 != 0:
        return False

    temp_text = re.sub(r"`[^`]*`", "", temp_text, flags=re.DOTALL)
    if temp_text.count("`") % 2 != 0:
        return False

    temp_text = re.sub(r'"[^"\\]*(\\.[^"\\]*)*"', "", temp_text)
    if temp_text.count('"') % 2 != 0:
        return False

    if temp_text.count("(") != temp_text.count(")"):
        return False
    if temp_text.count("[") != temp_text.count("]"):
        return False
    if temp_text.count("{") != temp_text.count("}"):
        return False

    depth = 0
    bracket_level = 0

    pattern = r"(\[|\]|\b(DEF|CLASS|ENUM|IF|ELSE\s+IF|ELSE|WHILE|FOR|TRY|SWITCH|ENDDEF|ENDCLASS|ENDENUM|ENDIF|ENDWHILE|ENDFOR|ENDTRY|ENDSWITCH)\b)"

    tokens = re.finditer(pattern, temp_text)

    start_keys = {"DEF", "CLASS", "ENUM", "IF", "WHILE", "FOR", "TRY", "SWITCH"}

    end_keys = {
        "ENDDEF",
        "ENDCLASS",
        "ENDENUM",
        "ENDIF",
        "ENDWHILE",
        "ENDFOR",
        "ENDTRY",
        "ENDSWITCH",
    }

    for match in tokens:
        token = match.group()

        if token in ("ELSE", "ELSE IF"):
            continue

        if token == "[":
            bracket_level += 1
        elif token == "]":
            bracket_level -= 1
        elif token in start_keys:
            depth += 1
        elif token in end_keys:
            depth -= 1

    return depth <= 0


def main():
    MAX_MEMORY_MB = 512
    MAX_INSTRUCTIONS = 100000

    set_memory_limit(MAX_MEMORY_MB)

    GLADLANG_VERSION = "0.1.9"
    GLADLANG_HELP = f"""
Usage: gladlang [command] [filename/code] [args...]

Commands:
  <no arguments>           Start the interactive GladLang shell.
  [filename.glad]          Execute a GladLang script file.
  ["code string"]          Execute inline GladLang code directly.
  [filename.glad] [args]   Execute script and pass args to INPUT().
  ["code string"] [args]   Execute inline code and pass args to INPUT().
  --help                   Show this help message and exit.
  --version                Show the interpreter version and exit.
"""

    if len(sys.argv) == 1:
        sys.stdout.write(f"Welcome to GladLang (v{GLADLANG_VERSION})\n")
        sys.stdout.write("Type 'exit' or 'quit' to close the shell.\n")
        sys.stdout.write("--------------------------------------------------\n")

        repl_context = Context("<repl>")
        repl_context.symbol_table = get_fresh_global_scope()

        full_text = ""

        while True:
            try:
                prompt = "GladLang > " if not full_text else "...        > "

                sys.stdout.write(prompt)
                sys.stdout.flush()

                line = sys.stdin.readline()

                if not line:
                    raise EOFError

                line = line.rstrip("\n")

                if not full_text and line.strip().lower() in ("exit", "quit"):
                    break

                full_text += line + "\n"

                if is_complete(full_text):
                    if full_text.strip() == "":
                        full_text = ""
                        continue

                    result, error = run(
                        "<stdin>",
                        full_text,
                        repl_context,
                        instruction_limit=MAX_INSTRUCTIONS,
                    )

                    if error:
                        sys.stdout.write(error.as_string() + "\n")
                    elif result:
                        clean_text = re.sub(r"#.*", "", full_text).strip()

                        if (
                            clean_text.startswith("LET ")
                            or clean_text.startswith("ENUM ")
                            or clean_text.startswith("FINAL ")
                        ):
                            pass
                        elif isinstance(result, Number) and result.value == 0:
                            pass
                        elif (
                            isinstance(result, List)
                            and len(result.elements) == 1
                            and result.elements[0].value == "NULL"
                        ):
                            pass
                        else:
                            sys.stdout.write(str(result) + "\n")

                    full_text = ""

            except KeyboardInterrupt:
                sys.stdout.write("\nKeyboardInterrupt\n")
                full_text = ""
                continue
            except MemoryError:
                sys.stdout.write("System Error: Memory Limit Exceeded\n")
                full_text = ""
            except EOFError:
                sys.stdout.write("\nExiting.\n")
                break
            except Exception as e:
                sys.stdout.write(f"Shell Error: {e}\n")
                full_text = ""

    elif len(sys.argv) >= 2:
        arg = sys.argv[1]

        if arg == "--help":
            sys.stdout.write(GLADLANG_HELP + "\n")

        elif arg == "--version":
            sys.stdout.write(f"GladLang v{GLADLANG_VERSION}\n")

        else:
            try:
                arg_input = arg
                script_args = sys.argv[2:]

                if script_args:
                    sys.stdin = io.StringIO("\n".join(script_args) + "\n")

                is_file = False
                try:
                    is_file = Path(arg_input).is_file()
                except OSError:
                    pass

                if is_file or arg_input.endswith(".glad"):
                    text = Path(arg_input).read_text(encoding="utf-8")
                    source_name = arg_input
                else:
                    text = arg_input
                    source_name = "<cmdline>"

                result, error = run(
                    source_name, text, instruction_limit=MAX_INSTRUCTIONS
                )

                if error:
                    sys.stderr.write(error.as_string() + "\n")

            except MemoryError:
                sys.stderr.write("System Error: Memory Limit Exceeded\n")
            except FileNotFoundError:
                sys.stderr.write(f"File not found: '{arg_input}'\n")
            except Exception as e:
                sys.stderr.write(f"An unexpected error occurred: {e}\n")

    else:
        sys.stdout.write("Error: Invalid arguments.\n")
        sys.stdout.write(GLADLANG_HELP + "\n")


if __name__ == "__main__":
    main()



from .constants import *
import codecs
from functools import lru_cache
from .errors import Position, IllegalCharError, InvalidSyntaxError


@lru_cache(maxsize=1024)
def decode_escapes(s):
    try:
        return codecs.decode(s, "unicode_escape")
    except:
        return s


class Token:
    def __init__(self, type_, value=None, pos_start=None, pos_end=None):
        self.type = type_
        self.value = value

        if pos_start:
            self.pos_start = pos_start.copy()
            self.pos_end = pos_start.copy()
            self.pos_end.advance()

        if pos_end:
            self.pos_end = pos_end.copy()

    def matches(self, type_, value):
        return self.type == type_ and self.value == value

    def __repr__(self):
        if self.value:
            return f"{self.type}:{self.value}"
        return f"{self.type}"


class Lexer:
    def __init__(self, fn, text):
        self.fn = fn
        self.text = text
        self.pos = Position(-1, 0, -1, fn, text)
        self.current_char = None
        self.advance()

    def advance(self):
        self.pos.advance(self.current_char)
        self.current_char = (
            self.text[self.pos.idx] if self.pos.idx < len(self.text) else None
        )

    def peek(self):
        peek_pos = self.pos.idx + 1
        if peek_pos < len(self.text):
            return self.text[peek_pos]
        return None

    def skip_comment(self):
        self.advance()

        while self.current_char != "\n" and self.current_char != None:
            self.advance()

    def make_tokens(self):
        tokens = []

        while self.current_char != None:
            if self.current_char in " \t\r\n":
                self.advance()
            elif self.current_char == "#":
                self.skip_comment()
            elif self.current_char in DIGITS:
                tok_or_tuple = self.make_number()
                if isinstance(tok_or_tuple, tuple):
                    return [], tok_or_tuple[1]
                tokens.append(tok_or_tuple)
            elif self.current_char.isidentifier():
                tokens.append(self.make_identifier())
            elif self.current_char == '"':
                tokens.append(self.make_string())
            elif self.current_char == "+":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "+":
                    tokens.append(
                        Token(GL_PLUSPLUS, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                elif self.current_char == "=":
                    tokens.append(
                        Token(GL_PLUSEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_PLUS, pos_start=pos_start))
            elif self.current_char == "-":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "-":
                    tokens.append(
                        Token(GL_MINUSMINUS, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                elif self.current_char == "=":
                    tokens.append(
                        Token(GL_MINUSEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_MINUS, pos_start=pos_start))
            elif self.current_char == "*":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "*":
                    tokens.append(Token(GL_POW, pos_start=pos_start, pos_end=self.pos))
                    self.advance()
                elif self.current_char == "=":
                    tokens.append(
                        Token(GL_MULEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_MUL, pos_start=pos_start))
            elif self.current_char == "/":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "/":
                    self.advance()
                    if self.current_char == "=":
                        tokens.append(
                            Token(GL_FLOORDIVEQ, pos_start=pos_start, pos_end=self.pos)
                        )
                        self.advance()
                    else:
                        tokens.append(
                            Token(GL_FLOORDIV, pos_start=pos_start, pos_end=self.pos)
                        )
                elif self.current_char == "=":
                    tokens.append(
                        Token(GL_DIVEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_DIV, pos_start=pos_start))
            elif self.current_char == "%":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "=":
                    tokens.append(
                        Token(GL_MODEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_MOD, pos_start=pos_start))
            elif self.current_char == "&":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "=":
                    tokens.append(
                        Token(GL_BIT_ANDEQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_BIT_AND, pos_start=pos_start))
            elif self.current_char == "|":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "=":
                    tokens.append(
                        Token(GL_BIT_OREQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_BIT_OR, pos_start=pos_start))
            elif self.current_char == "~":
                tokens.append(Token(GL_BIT_NOT, pos_start=self.pos))
                self.advance()
            elif self.current_char == "^":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "=":
                    tokens.append(
                        Token(GL_BIT_XOREQ, pos_start=pos_start, pos_end=self.pos)
                    )
                    self.advance()
                else:
                    tokens.append(Token(GL_BIT_XOR, pos_start=pos_start))
            elif self.current_char == "<":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == "<":
                    self.advance()
                    if self.current_char == "=":
                        tokens.append(
                            Token(GL_LSHIFTEQ, pos_start=pos_start, pos_end=self.pos)
                        )
                        self.advance()
                    else:
                        tokens.append(
                            Token(GL_LSHIFT, pos_start=pos_start, pos_end=self.pos)
                        )
                elif self.current_char == "=":
                    tokens.append(Token(GL_LTE, pos_start=pos_start, pos_end=self.pos))
                    self.advance()
                else:
                    tokens.append(Token(GL_LT, pos_start=pos_start))
            elif self.current_char == ">":
                pos_start = self.pos.copy()
                self.advance()
                if self.current_char == ">":
                    self.advance()
                    if self.current_char == "=":
                        tokens.append(
                            Token(GL_RSHIFTEQ, pos_start=pos_start, pos_end=self.pos)
                        )
                        self.advance()
                    else:
                        tokens.append(
                            Token(GL_RSHIFT, pos_start=pos_start, pos_end=self.pos)
                        )
                elif self.current_char == "=":
                    tokens.append(Token(GL_GTE, pos_start=pos_start, pos_end=self.pos))
                    self.advance()
                else:
                    tokens.append(Token(GL_GT, pos_start=pos_start))
            elif self.current_char == "(":
                tokens.append(Token(GL_LPAREN, pos_start=self.pos))
                self.advance()
            elif self.current_char == ")":
                tokens.append(Token(GL_RPAREN, pos_start=self.pos))
                self.advance()
            elif self.current_char == ",":
                tokens.append(Token(GL_COMMA, pos_start=self.pos))
                self.advance()
            elif self.current_char == ".":
                tokens.append(Token(GL_DOT, pos_start=self.pos))
                self.advance()
            elif self.current_char == "[":
                tokens.append(Token(GL_LSQUARE, pos_start=self.pos))
                self.advance()
            elif self.current_char == "]":
                tokens.append(Token(GL_RSQUARE, pos_start=self.pos))
                self.advance()
            elif self.current_char == "!":
                tok, error = self.make_not_equals()
                if error:
                    return [], error
                tokens.append(tok)
            elif self.current_char == "=":
                tokens.append(self.make_equals())
            elif self.current_char == "`":
                tokens += self.make_template_string()
            elif self.current_char == "<":
                tokens.append(self.make_less_than())
            elif self.current_char == ">":
                tokens.append(self.make_greater_than())
            elif self.current_char == "{":
                tokens.append(Token(GL_LBRACE, pos_start=self.pos))
                self.advance()
            elif self.current_char == "}":
                tokens.append(Token(GL_RBRACE, pos_start=self.pos))
                self.advance()
            elif self.current_char == ":":
                tokens.append(Token(GL_COLON, pos_start=self.pos))
                self.advance()
            elif self.current_char == "?":
                tokens.append(Token(GL_QMARK, pos_start=self.pos))
                self.advance()
            else:
                pos_start = self.pos.copy()
                char = self.current_char
                self.advance()
                return [], IllegalCharError(pos_start, self.pos, "'" + char + "'")

        tokens.append(Token(GL_EOF, pos_start=self.pos))
        return tokens, None

    def make_number(self):
        num_str = ""
        dot_count = 0
        pos_start = self.pos.copy()

        if self.current_char == "0":
            peek_char = self.peek()

            if peek_char in ("x", "X"):
                self.advance()
                self.advance()

                if self.current_char is None or not (
                    self.current_char in DIGITS or self.current_char.lower() in "abcdef"
                ):
                    return Token(GL_INT, 0, pos_start, self.pos), IllegalCharError(
                        pos_start, self.pos, "Invalid hex literal"
                    )

                while self.current_char is not None and (
                    self.current_char in DIGITS or self.current_char.lower() in "abcdef"
                ):
                    num_str += self.current_char
                    self.advance()
                return Token(GL_INT, int(num_str, 16), pos_start, self.pos)

            elif peek_char in ("o", "O"):
                self.advance()
                self.advance()

                if self.current_char is None or self.current_char not in "01234567":
                    return Token(GL_INT, 0, pos_start, self.pos), IllegalCharError(
                        pos_start, self.pos, "Invalid octal literal"
                    )

                while self.current_char is not None and self.current_char in "01234567":
                    num_str += self.current_char
                    self.advance()
                return Token(GL_INT, int(num_str, 8), pos_start, self.pos)

            elif peek_char in ("b", "B"):
                self.advance()
                self.advance()

                if self.current_char is None or self.current_char not in "01":
                    return Token(GL_INT, 0, pos_start, self.pos), IllegalCharError(
                        pos_start, self.pos, "Invalid binary literal"
                    )

                while self.current_char is not None and self.current_char in "01":
                    num_str += self.current_char
                    self.advance()
                return Token(GL_INT, int(num_str, 2), pos_start, self.pos)

        while self.current_char != None and self.current_char in DIGITS + ".":
            if self.current_char == ".":
                if dot_count == 1:
                    break
                dot_count += 1
                num_str += "."
            else:
                num_str += self.current_char
            self.advance()

        if dot_count == 0:
            return Token(GL_INT, int(num_str), pos_start, self.pos)
        else:
            return Token(GL_FLOAT, float(num_str), pos_start, self.pos)

    def make_string(self, quote_type):
        string_content = ""
        escape_character = False
        self.advance()

        while self.current_char != None and (
            self.current_char != quote_type or escape_character
        ):
            if escape_character:
                string_content += "\\" + self.current_char
                escape_character = False
            else:
                if self.current_char == "\\":
                    escape_character = True
                else:
                    string_content += self.current_char
            self.advance()

        self.advance()

        string_content = decode_escapes(string_content)

        return Token(GL_STRING, string_content, self.pos_start, self.pos)

    def make_identifier(self):
        id_str = ""
        pos_start = self.pos.copy()

        while self.current_char != None and (id_str + self.current_char).isidentifier():
            id_str += self.current_char
            self.advance()

        tok_type = GL_KEYWORD if id_str in KEYWORDS else GL_IDENTIFIER
        return Token(tok_type, id_str, pos_start, self.pos)

    def make_not_equals(self):
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == "=":
            self.advance()
            return Token(GL_NE, pos_start=pos_start, pos_end=self.pos), None

        self.advance()
        return None, InvalidSyntaxError(pos_start, self.pos, "Expected '=' after '!'")

    def make_equals(self):
        tok_type = GL_EQ
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == "=":
            self.advance()
            tok_type = GL_EE

        return Token(tok_type, pos_start=pos_start, pos_end=self.pos)

    def make_template_string(self):
        tokens = []
        pos_start = self.pos.copy()
        self.advance()

        tokens.append(Token(GL_LPAREN, pos_start=pos_start))

        string_part = ""
        escape_character = False

        while self.current_char != None and (
            self.current_char != "`" or escape_character
        ):

            if not escape_character and self.current_char == "$" and self.peek() == "{":
                tokens.append(Token(GL_STRING, string_part, pos_start=pos_start))
                string_part = ""

                tokens.append(Token(GL_PLUS, pos_start=self.pos))
                tokens.append(Token(GL_IDENTIFIER, "STR", pos_start=self.pos))
                tokens.append(Token(GL_LPAREN, pos_start=self.pos))

                self.advance()
                self.advance()

                expr_str = ""
                brace_count = 1
                while self.current_char != None and brace_count > 0:
                    if self.current_char == "{":
                        brace_count += 1
                    elif self.current_char == "}":
                        brace_count -= 1

                    if brace_count > 0:
                        expr_str += self.current_char
                        self.advance()

                sub_lexer = Lexer(self.fn, expr_str)
                sub_tokens, error = sub_lexer.make_tokens()

                if sub_tokens and sub_tokens[-1].type == GL_EOF:
                    sub_tokens.pop()

                tokens.extend(sub_tokens)

                tokens.append(Token(GL_RPAREN, pos_start=self.pos))
                tokens.append(Token(GL_PLUS, pos_start=self.pos))

                self.advance()

            elif escape_character:
                if self.current_char == "n":
                    string_part += "\n"
                elif self.current_char == "t":
                    string_part += "\t"
                elif self.current_char == "r":
                    string_part += "\r"
                elif self.current_char == "`":
                    string_part += "`"
                elif self.current_char == "\\":
                    string_part += "\\"
                elif self.current_char == '"':
                    string_part += '"'
                elif self.current_char == "'":
                    string_part += "'"
                elif self.current_char == "$":
                    string_part += "$"
                else:
                    string_part += self.current_char

                escape_character = False
                self.advance()

            elif self.current_char == "\\":
                escape_character = True
                self.advance()

            else:
                string_part += self.current_char
                self.advance()

        tokens.append(Token(GL_STRING, string_part, pos_start=pos_start))
        tokens.append(Token(GL_RPAREN, pos_start=self.pos))

        self.advance()
        return tokens

    def make_string(self):
        string = ""
        pos_start = self.pos.copy()

        is_multiline = False

        self.advance()

        if self.current_char == '"' and self.peek() == '"':
            is_multiline = True
            self.advance()
            self.advance()

        escape_character = False

        while self.current_char != None:
            if escape_character:
                if self.current_char == "n":
                    string += "\n"
                elif self.current_char == "t":
                    string += "\t"
                elif self.current_char == '"':
                    string += '"'
                elif self.current_char == "\\":
                    string += "\\"
                else:
                    string += self.current_char
                escape_character = False
            elif self.current_char == "\\":
                escape_character = True
            elif self.current_char == '"':
                if is_multiline:
                    if self.peek() == '"':
                        self.advance()
                        if self.peek() == '"':
                            self.advance()
                            break
                        else:
                            string += '""'
                    else:
                        string += '"'
                else:
                    break
            else:
                string += self.current_char

            self.advance()

        self.advance()
        return Token(GL_STRING, string, pos_start, self.pos)

    def make_less_than(self):
        tok_type = GL_LT
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == "=":
            self.advance()
            tok_type = GL_LTE

        return Token(tok_type, pos_start=pos_start, pos_end=self.pos)

    def make_greater_than(self):
        tok_type = GL_GT
        pos_start = self.pos.copy()
        self.advance()

        if self.current_char == "=":
            self.advance()
            tok_type = GL_GTE

        return Token(tok_type, pos_start=pos_start, pos_end=self.pos)

    def make_plus_equals(self):
        token_type = GL_PLUS
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_PLUSEQ
        return Token(token_type, pos_start=self.pos)

    def make_minus_equals(self):
        token_type = GL_MINUS
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_MINUSEQ
        return Token(token_type, pos_start=self.pos)

    def make_mul_equals(self):
        token_type = GL_MUL
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_MULEQ
        return Token(token_type, pos_start=self.pos)

    def make_div_equals(self):
        token_type = GL_DIV
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_DIVEQ
        return Token(token_type, pos_start=self.pos)

    def make_pow_equals(self):
        token_type = GL_POW
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_POWEQ
        return Token(token_type, pos_start=self.pos)

    def make_mod_equals(self):
        token_type = GL_MOD
        self.advance()
        if self.current_char == "=":
            self.advance()
            token_type = GL_MODEQ
        return Token(token_type, pos_start=self.pos)
